
from typing import (
    Any,
    Union, Optional,
    Tuple, List, Set, Sequence, Dict,
)

import asyncio
import httpx
import json
import datetime
from http.client import responses as RESPONSE_CODES

from .models import Message, System, Member, Switch, Timestamp
from .errors import *
from .utils import *
from . import endpoint

class Client:
    """Represents a client that interacts with the PluralKit API.

    Args:
        token: The PluralKit authorization token, received by the ``pk;token`` command.

    Keyword args:
        async_mode: Whether the client runs asynchronously (``True``, default) or not (``False``).
        user_agent: The User-Agent header to use with the API.

    Attributes:
        token (Optional[str]): The client's PluralKit authorization token.
        async_mode (bool): Whether the client runs asynchronously (``True``) or not (``False``).
        user_agent (Optional[str]): The User-Agent header used with the API.
        headers (Dict[str,str]): 
        content_headers (Dict[str,str]): 
    """

    SERVER = "https://api.pluralkit.me/v1"

    def __init__(self, token: Optional[str]=None, *,
        async_mode: bool=True,
        user_agent: Optional[str]=None
    ):
        self.async_mode = async_mode
        self.token = token
        self.headers = {}
        if token:
            self.headers["Authorization"] = token
        self.user_agent = user_agent
        if user_agent:
            self.headers["User-Agent"] = user_agent
        self.content_headers = self.headers.copy()
        self.content_headers['Content-Type'] = "application/json"
        self._id = None

    @property
    def id(self) -> Optional[str]:
        """The five-letter lowercase PluralKit system ID of this client, if initialized.
        """
        return self._id

    async def _check_self_id(self):
        if self._id is None:
            system = await self._get_system()
            self._id = system.id
    
    @staticmethod
    def _get_system_url(self, system) -> str:
        if system is None:
            if not self.token: raise AuthorizationError() # please pass in your token to the client
            # get own system
            url = f"{self.SERVER}/s"
        elif isinstance(system, System):
            # System object
            url = f"{self.SERVER}/s/{system.id}"
        elif isinstance(system, str):
            # system ID
            url = f"{self.SERVER}/s/{system}"
        elif isinstance(system, int):
            # Discord user ID
            url = f"{self.SERVER}/a/{system}"

        return url

    @endpoint.func
    async def get_system(self, system: Union[System,str,int,None]=None) -> System:
        """Return a system by its system ID or Discord user ID.

        Args:
            system: The system ID (as `str`), Discord user ID (as `int`), or `~v1.models.System`
                object of the system. If ``None``, returns the system of the client.

        Returns:
            System: The retrieved system.
        """
        url = Client._get_system_url(self, system)


        async with httpx.AsyncClient(headers=self.headers) as session:
            response = await session.get(url)
            if response.status_code == 401:
                raise AuthorizationError()
            elif response.status_code == 404:
                if isinstance(system, str):
                    raise SystemNotFound(system)
                elif isinstance(system, int):
                    raise DiscordUserNotFound(system)
                
                if response.status_code != 200: # catch-all
                    raise HTTPError(response.status_code)

            resp = response.json()

            system = System.from_json(resp)

            if url.endswith("/s"):
                # remember self ID for the future
                self._id = system.id

            return system
    
    @endpoint.func
    async def edit_system(self, **kwargs) -> System:
        """"Edits one's own system
        
        Note:
            Any fields not passed will retain their prior value.
            Any fields passed with null will clear or reset to the default.
            
        Args:
            **kwargs: Any number of keyworded patchable values from `PluralKit's system model`_.
            
        Keyword Args:
            name(str): The new system name.
            description(Optional[str]): The new system description.
            tag(Optional[str]): The new system tag.
            avatar_url(Optional[str]): The new system avatar.
            tz(Optional[str]): The timezone as a tzdb identifier, null will store "UTC"
            description_privacy(Optional[str]): The new system description privacy value
            member_list_privacy(Optional[str]): The new system member list privacy value
            front_privacy(Optional[str]): The new system front privacy value
            metadata_privacy(Optional[str]): The new system metadata privacy value

        Returns:
            System: The updated system.

        .. _`PluralKit's system model`: https://pluralkit.me/api/#system-model
        """

        if self.token is None:
            raise AuthorizationError()

        url = f"https://api.pluralkit.me/v1/s"
    
        for key, value in kwargs.items():
            kwargs = await system_value(key=key, value=value)
        
        json_object = json.dumps(kwargs, ensure_ascii=False)

        async with httpx.AsyncClient(headers=self.content_headers) as session:
            response = await session.patch(url, data=json_object)
            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)

            resp = response.json()

            system = System.from_json(resp)

            return system

    @endpoint.func
    async def get_fronters(self, system=None) -> Tuple[Timestamp, List[Member]]:
        """Fetches the current fronters of a system.
        
        Args:
            system(Optional[Union[str, System]]): The system to fetch fronters from.
        
        Returns:
            Set(Timestamp, List[Member]): A set containing a Timestamp object and a list of current
            fronters in Member objects
        """
        
        if system is None: 
            await self._check_self_id()
            url = f"{self.SERVER}/s/{self.id}/fronters"
        elif isinstance(system, str):
            url = f"{self.SERVER}/s/{system}/fronters"
        elif isinstance(system, System):
            url = f"{self.SERVER}/s/{system.id}/fronters"
        
        async with httpx.AsyncClient(headers=self.headers) as session:
            response = await session.get(url)
            if response.status_code == 401:
                raise AuthorizationError()
            elif response.status_code == 403:
                raise AccessForbidden()
            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)
            else:
                resp = response.json()
                member_list = []
                for fronter in resp['members']:
                    member_list.append(Member.from_json(fronter))
                timestamp = Timestamp.from_json(resp['timestamp'])
                return (timestamp, member_list)

    @endpoint.iter
    async def get_members(self, system: Union[System,str,int,None]=None):
        """Retrieve list of a system's members.

        Args:
            system: The system ID, Discord user ID, or System object of the system. If ``None``,
                returns a list of the client system's members.

        Yields:
            Member: The next system member.
        """

        if system is None:
            # get own system
            await self._check_self_id()
            url = f"{self.SERVER}/s/{self.id}/members"
        elif isinstance(system, System):
            # System object
            url = f"{self.SERVER}/s/{system.id}/members"
        elif isinstance(system, str):
            # system ID
            url = f"{self.SERVER}/s/{system}/members"
        elif isinstance(system, int):
            # Discord user ID
            system = await self._get_system(system)
            url = f"{self.SERVER}/s/{system.id}/members"

        async with httpx.AsyncClient(headers=self.headers) as session:
            response = await session.get(url)
            if response.status_code == 401:
                raise AuthorizationError()
            elif response.status_code == 403:
                raise AccessForbidden()
            elif response.status_code == 404:
                if isinstance(system, str):
                    raise SystemNotFound(system)
                elif isinstance(system, int):
                    raise DiscordUserNotFound(system)
                
            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)

            resp = response.json()

            for item in resp:
                member = Member.from_json(item)

                yield member
    
    @endpoint.func
    async def get_member(self, member_id: str) -> Member:
        """Gets a system member.

        Args:
            member_id: The ID of the member to be fetched.

        Returns:
            Member: The member with the given ID.

        .. _`PluralKit's member model`: https://pluralkit.me/api/#member-model
        """
        async with httpx.AsyncClient(headers=self.headers) as session:
            response = await session.get(f"{self.SERVER}/m/{member_id}")
            if response.status_code == 401:
                raise AuthorizationError()
            elif response.status_code == 200:
                item = response.json()
                return Member.from_json(item)
            else:
                raise Exception(f"Something went wrong with your request. You received a "
                        f"{response.status_code} http code, here is a list of possible http codes "
                        "https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#4xx_client_errors")

    @endpoint.func
    async def new_member(self, **kwargs) -> Member:
        """Creates a new member of one's system.

        Note:
            The system's `authorization token`_ must be set in order to use :meth:`new_member`.

        Args:
            name: Name of the new member
            **kwargs: Any number of keyworded patchable values from `PluralKit's member model`_.
        
        Keyword Args:
            display_name (Optional[str]): New display name of the member. Default is None.
            description (Optional[str]): New description of the member. Default is None.
            pronouns (Optional[str]): New pronouns of the member. Default is None.
            color (Union[Color,str,None]): New color of the member. If a string, must be formatted
                as a 6-character hex string (e.g. ``"ff7000"``), sans the # symbol. Default is
                ``None``.
            avatar_url (Optional[str]): New avatar URL for the member. Default is None.
            birthday (Union[Timestamp,str]): New birthdate of the member. If a string, must be
                formatted as ``YYYY-MM-DD``, in which case, year of ``0001`` or ``0004`` represents
                a hidden year. Default is None.
            proxy_tags (Union[ProxyTags,Sequence[ProxyTag],Sequence[Dict[str,str]]]): New proxy
                tags of the member. May be a `~v1.models.ProxyTags` object, a sequence of
                `~v1.models.ProxyTag` objects, or a sequence of Python dictionaries with the
                keys "prefix" and "suffix". Default is an empty set of proxy tags.
            keep_proxy (bool): New truth value for whether to display the member's proxy tags in
                the proxied message. Default is ``False``.
            visibility (Union[Privacy,str,None]): New visibility privacy for the member. Must be
                either :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`.
                Default is public.
            name_privacy (Union[Privacy,str,None]): New name privacy for the member. Must be either
                :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`. Default is
                public.
            description_privacy (Union[Privacy,str,None]): New description privacy for the member.
                Must be either :attr:`~v1.models.Privacy.PUBLIC` or
                :attr:`~v1.models.Privacy.PRIVATE`. Default is public.
            avatar_privacy (Union[Privacy,str,None]): New avatar privacy for the member. Must be
                either :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`.
                Default is public.
            pronoun_privacy (Union[Privacy,str]): New pronouns privacy for the member. Must be
                either :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`.
                Default is public.
            metadata_privacy (Union[Privacy,str]): New metadata (eg. creation timestamp, message
                count, etc.) privacy for the member. Must be either
                :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`. Default is
                public.

        Returns:
            Member: The new member.

        .. _`PluralKit's member model`: https://pluralkit.me/api/#member-model
        .. _`authorization token`: https://pluralkit.me/api/#authentication
        """
        #Finish editing this so it makes sense in the context it's in
        if self.token is None:
            raise AuthorizationError()
        self._name = None
        for key, value in kwargs.items():
            if key == "name":
                self._name = value
            kwargs = await member_value(kwargs=kwargs, key=key, value=value)
        if self._name is None:
            raise Exception("Must have field 'name'")

        json_payload = json.dumps(kwargs, ensure_ascii=False)
        async with httpx.AsyncClient(headers=self.content_headers) as session:
            response = await session.post(f"{self.SERVER}/m/", data=json_payload)
            if response.status_code == 401:
                raise AuthorizationError
            elif response.status_code == 200:
                item = response.json()
                return Member.from_json(item)
            else:
                raise Exception(f"Something went wrong with your request. You received a "
                        f"{response.status_code} http code, here is a list of possible http codes "
                        "https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#4xx_client_errors")

    @endpoint.func
    async def edit_member(self, member_id: str, **kwargs) -> Member:
        """Edits a member of one's system.

        Note:
            The system's `authorization token`_ must be set in order to use :meth:`edit_member`.

        Args:
            member_id: The ID of the member to be edited.
            **kwargs: Any number of keyworded patchable values from `PluralKit's member model`_.
        
        Keyword Args:
            name (str): New name of the member.
            display_name (Optional[str]): New display name of the member. If ``None`` is passed,
                this field is cleared.
            description (Optional[str]): New description of the member. If ``None`` is passed, this
                field is cleared.
            pronouns (Optional[str]): New pronouns of the member. If ``None`` is passed, this field
                is cleared.
            color (Union[Color,str,None]): New color of the member. If a string, must be formatted
                as a 6-character hex string (e.g. ``"ff7000"``), sans the # symbol. If ``None`` is
                passed, this field is cleared.
            avatar_url (Optional[str]): New avatar URL for the member. If ``None`` is passed, this
                field is cleared.
            birthday (Union[Birthday,str]): New birthdate of the member. If a string, must be
                formatted as ``YYYY-MM-DD``, in which case, a year of ``0001`` or ``0004``
                represents a hidden year. If ``None`` is passed, this field is cleared.
            proxy_tags (Union[ProxyTags,Sequence[ProxyTag],Sequence[Dict[str,str]]]): New proxy
                tags of the member. May be a `~v1.models.ProxyTags` object, a sequence of
                `~v1.models.ProxyTag` objects, or a sequence of Python dictionaries with the
                keys "prefix" and "suffix".
            keep_proxy (bool): New truth value for whether to display the member's proxy tags in
                the proxied message.
            visibility (Union[Privacy,str,None]): New visibility privacy for the member. Must be
                either :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`. If
                ``None`` is passed, this field is reset to public.
            name_privacy (Union[Privacy,str,None]): New name privacy for the member. Must be either
                :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`. If
                ``None`` is passed, this field is reset to public.
            description_privacy (Union[Privacy,str,None]): New description privacy for the member.
                Must be either :attr:`~v1.models.Privacy.PUBLIC` or
                :attr:`~v1.models.Privacy.PRIVATE`. If ``None`` is passed, this field is reset to
                public.
            avatar_privacy (Union[Privacy,str,None]): New avatar privacy for the member. Must be
                either :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`. If
                ``None`` is passed, this field is reset to public.
            pronoun_privacy (Union[Privacy,str]): New pronouns privacy for the member. Must be
                either :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`. If
                ``None`` is passed, this field is reset to public.
            metadata_privacy (Union[Privacy,str]): New metadata (eg. creation timestamp, message
                count, etc.) privacy for the member. Must be either
                :attr:`~v1.models.Privacy.PUBLIC` or :attr:`~v1.models.Privacy.PRIVATE`. If
                ``None`` is passed, this field is reset to public.

        Returns:
            Member: The updated member.

        .. _`PluralKit's member model`: https://pluralkit.me/api/#member-model
        .. _`authorization token`: https://pluralkit.me/api/#authentication
        """

        if self.token is None:
            raise AuthorizationError()
        
        for key, value in kwargs.items():
            kwargs = await member_value(kwargs=kwargs, key=key, value=value)

        json_payload = json.dumps(kwargs, ensure_ascii=False)
        async with httpx.AsyncClient(headers=self.content_headers) as session:
            response = await session.patch(f"{self.SERVER}/m/{member_id}", data=json_payload)
            if response.status_code == 401:
                raise AuthorizationError
            elif response.status_code == 200:
                item = response.json()
                return Member.from_json(item)
            else:
                raise Exception(f"Something went wrong with your request. You received a "
                        f"{response.status_code} http code, here is a list of possible http codes "
                        "https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#4xx_client_errors")
        
    @endpoint.func
    async def delete_member(self, member_id) -> None:
        """Deletes a member of one's system
        
        Note:
            The system's `authorization token`_ must be set in order to use :meth:`delete_member`.
        
        Args:
            member_id (str): The ID of the member to be deleted.

        .. _`authorization token`: https://pluralkit.me/api/#authentication
        """

        url = f"{self.SERVER}/m/{member_id}"

        async with httpx.AsyncClient(headers=self.headers) as session:
            response = await session.delete(url)
            if response.status_code == 401:
                if response.status_code == 401:
                    raise AuthorizationError()
                elif response.status_code == 403:
                    raise AccessForbidden()

                if response.status_code != 204: # catch-all
                    raise HTTPError(response.status_code)

    @endpoint.iter
    async def get_switches(self, system=None):
        """Fetches the switch history of a system.
        
        Args:
            system (Optional[Union[str, System]])
            
        Yields:
             Switch: The next switch.
        """
        
        if system is None:
            #Authorized system
            await self._check_self_id()
            system_url = f"/s/{self.id}"
        elif isinstance(system, System):
            # System object
            system_url = f"/s/{system.id}"
        elif isinstance(system, str):
            # system ID
            system_url = f"/s/{system}"

        url = f"{self.SERVER}{system_url}/switches"

        async with httpx.AsyncClient(headers=self.headers) as session:
            response = await session.get(url)
            if response.status_code == 401:
                raise AuthorizationError()
            elif response.status_code == 403:
                raise AccessForbidden()
            elif response.status_code == 404:
                if isinstance(system, str):
                    raise SystemNotFound(system)
                elif isinstance(system, int):
                    raise DiscordUserNotFound(system)
                
            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)

            resp = response.json()

            for item in resp:
                switch = Switch.from_json(item)
                yield switch

    @endpoint.func
    async def new_switch(self, members) -> Switch:
        """Creates a new switch
        
        Note:
            The system's `authorization token`_ must be set in order to use :meth:`delete_member`.
        
        Args:
            members(Sequence[str]): A list of members that will be present in the new switch.

        Returns:
            Switch: The newly created switch.

        .. _`authorization token`: https://pluralkit.me/api/#authentication
        """

        if self.token is None:
            raise AuthorizationError()
        
        url = f"{self.SERVER}/s/switches"
        payload = json.dumps({"members": members}, ensure_ascii=False)

        async with httpx.AsyncClient(headers=self.content_headers) as session:
            response = await session.post(url, data=payload)
            if response.status_code == 401:
                raise AuthorizationError()
            elif response.status_code == 403:
                raise AccessForbidden()
                
            if response.status_code != 204: # catch-all
                raise HTTPError(response.status_code)
    
    @endpoint.func
    async def get_message(self, message: Union[str, int, Message]) -> Message:
        """Fetches a message proxied by pluralkit
        
        Note:
            Messages proxied by pluralkit can be fetched either with the proxy message's id, or the
            id of the original message that triggered the proxy.
        
        Args:
            Message: The message to be fetched.
        """
        
        if isinstance(message, (str, int)):
            url = f"{self.SERVER}/msg/{message}"
        elif isinstance(message, Message):
            url = f"{self.SERVER}/msg/{message.id}"

        async with httpx.AsyncClient(headers=self.headers) as session:
            response = await session.get(url)
            if response.status_code == 401:
                raise AuthorizationError()
            elif response.status_code == 403:
                raise AccessForbidden()
            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)
            else:
                resp = response.json()
                
                return Message.from_json(resp)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.append(os.path.abspath(".."))
sys.path.append(os.path.abspath("../pluralkit/v1"))
sys.path.insert(0, os.path.abspath(".."))
from pluralkit import __version__
from datetime import datetime

# -- Project information -----------------------------------------------------

project = "pluralkit.py"
copyright = "2021, Madison Landry, Alyx Warner"
author = "Madison Landry, Alyx Warner"

# The full version
release = __version__

print(f"Building docs for: {project} v{release}")

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
   "sphinx.ext.napoleon",

   "sphinx.ext.intersphinx",
   "sphinx.ext.autodoc",
   "sphinx_autodoc_typehints"
]
intersphinx_mapping = {
   "python": ("https://docs.python.org/3", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = "alabaster"
#html_theme = "basicstrap" # https://pythonhosted.org/sphinxjp.themes.basicstrap/
html_theme = "sphinx_book_theme" # https://sphinx-themes.org/sample-sites/sphinx-book-theme/

today = datetime.utcnow().strftime(
   "<time datetime=\"%Y-%m-%dT%H:%MZ\">" \
   "%b %d at %H:%M UTC"
   "</time>"
)
html_theme_options = {
   "home_page_in_toc": True,
   "toc_title": "Jump to",
   "repository_url": "https://github.com/almonds0166/pluralkit.py",
   "path_to_docs": "docs/",
   "use_repository_button": True,
   #"use_edit_page_button": True,
   "extra_navbar": f"<p>Last built {today}</p>", # extra footer on left sidebar
}

html_title = f"{project} v{release} docs"
html_logo = "source/images/pluralkit_py_logo.png"
html_favicon = "source/images/pluralkit_py_logo.png"
show_navbar_depth = 3

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

default_role = "py:obj" # so we can type `blah` instead of :class:`blah`



from typing import (
    Any,
    Union, Optional,
    Tuple, List, Set, Sequence, Dict,
)
import datetime
from http.client import responses as RESPONSE_CODES

import httpx
import colour
import pytz

from .models import ProxyTag, Privacy, Timezone, Color
from .errors import *

MEMBER_ATTRS = (
    "name", 
    "display_name", 
    "description", 
    "pronouns", 
    "color", 
    "avatar_url", 
    "birthday", 
    "proxy_tags", 
    "keep_proxy", 
    "visibility", 
    "name_privacy", 
    "description_privacy", 
    "avatar_privacy", 
    "pronoun_privacy",
    "metadata_privacy",
    "birthday_privacy"
)

SYSTEM_ATTRS = (
    "name",
    "description",
    "tag",
    "avatar_url",
    "tz",
    "description_privacy",
    "member_list_privacy",
    "front_privacy",
    "front_history_privacy"
)

async def member_value(kwargs, key, value):
    """Prepares the kwargs given to `~v1.client.Client` methods for PluralKit's API, for internal
    use.
    """
    if not key in MEMBER_ATTRS:
        raise InvalidKwarg(key)
    if key == "color":
        if value is not None:
            kwargs[key] = Color.parse(value).hex_l[1:]
    elif key == "birthday":
        if isinstance(value, datetime.date): # will catch Timestamp and Birthday objects
            kwargs[key] = value.strftime("%Y-%m-%d")
        elif isinstance(value, str):
            try:
                datetime.datetime.strptime(value, "%Y-%m-%d")
            except:
                raise InvalidBirthday(value)
    elif key == "keep_proxy":
        if not isinstance(value, bool):
            raise ValueError(
                f"Keyword arg `keep_proxy` must be a boolean value; received {type(key)=}."
            )
    elif key in MEMBER_ATTRS [9:]:
        if not value in ("public", "private", None) and not value in Privacy:
            raise ValueError(
                f"Keyword arg `{key}` must be in (None, 'public', 'private') or a Privacy; " \
                f"instead was {value=}."
            )
        if value in Privacy:
            kwargs[key] = value.value # convert Privacy enum to strt
    elif key == "avatar_url":
        async with httpx.AsyncClient() as session:
            response = session.head(value)
            code = response.status_code
            if code != 200:
                raise ValueError(
                    f"Invalid URL passed. Received {code} {RESPONSE_CODES[code]}."
                )
    elif key == "proxy_tags":
        proxy_tags = []
        for proxy_tag in value:
            if isinstance(proxy_tag, ProxyTag):
                proxy_tags.append(proxy_tag.json()) # convert to dict
            elif isinstance(proxy_tag, dict):
                proxy_tags.append(proxy_tag)
            else:
                raise ValueError(
                    f"Keyword arg `proxy_tags` must be a ProxyTags object, a sequence of " \
                    f"ProxyTag objects, or a sequence of dict containing the keys 'prefix' " \
                    f"and 'suffix'."
                    )
        kwargs[key] = proxy_tags

    return kwargs

async def system_value(key, value):
    if not key in SYSTEM_ATTRS:
        raise InvalidKwarg(key)
    if key in ("name", "description", "tag", "avatar_url") and not isinstance(value, str):
        raise ValueError(f"{key}'s value must be of type string")
    if key == "tz" and not isinstance(value, (Timezone, str)):
        raise ValueError(f"{key}'s value must be of type string or Timezone")
    if key in ("description_privacy", "member_list_privacy", "front_privacy", 
              "front_history_privacy") and not isinstance(value, bool):
        raise ValueError(f"{key}'s value must be of type bool")




import logging
import sys
sys.path.append("..")
from random import choice

import asyncio

from pluralkit import Client, Member

choices = {
    "id": ["irqkk", "cewel"],
    "name": ["test_1", "test_2", "test_3"],
    "display_name": ["testing_1", "testing_2", "testing_3"],
    "description": ["testing..._1", "testing..._2", "testing..._3"],
    "pronouns": ["they/them", "she/her", "he/him"],
    "color": ["red", "#ffffff", ""],
    "created": ["2019-01-01T15:00:00.654321Z", "2019-01-01T15:00:00.654321Z"],
    "avatar_url": ["https://cdn.pixabay.com/photo/2015/04/23/22/00/tree-736885__480.jpg", "https://image.shutterstock.com/image-photo/mountains-under-mist-morning-amazing-260nw-1725825019.jpg"],
    "birthday": ["1904-02-28", "2025-06-12", "0834-12-23"],
    "proxy_tags": [[{"prefix": "t;"}], [{"suffix": ";t"}], [{"prefix": "te;"}]],
    "keep_proxy": [True, False],
    "visibility": ["public", "private", None],
    "name_privacy": ["public", "private", None],
    "description_privacy": ["public", "private", None],
    "avatar_privacy": ["public", "private", None],
    "birthday_privacy": ["public", "private", None],
    "pronoun_privacy": ["public", "private", None],
    "metadata_privacy": ["public", "private", None],
}

async def main(pk, member):
    m = Member.from_dict(member)
    await pk.edit_member(member_id=m.id, name=m.name, display_name=m.display_name, description=m.description, pronouns=m.pronouns, color=m.color, avatar_url=m.avatar_url, birthday=m.birthday, proxy_tags=m.proxy_tags, keep_proxy=m.keep_proxy, visibility=m.visibility, name_privacy=m.name_privacy, description_privacy=m.description_privacy, avatar_privacy=m.avatar_privacy, birthday_privacy=m.birthday_privacy, pronoun_privacy=m.pronoun_privacy, metadata_privacy=m.metadata_privacy)
    
def run(token: str=None):
    if not token:
        print("Enter the token, if desired, otherwise press Enter.")
        token = input("> ").strip() # getpass doesn't allow copy-paste
    pk = Client(token)
    try:
        member = {}
        for key, value in choices.items():
            i = choice(value)
            member[key] = i
        asyncio.run(main(pk, member))
    except KeyboardInterrupt:
        print("^C")
    except:
        logging.exception("An exception occurred")

if __name__ == "__main__":
    run()


import logging
import sys
sys.path.append("..")

import asyncio

from pluralkit import Client

async def main(pk, id):
    async for member in pk.get_members(id):
        print(f"{member.name} (`{member.id}`)")

def run(token: str=None):
    if not token:
        print("Enter the token, if desired, otherwise press Enter.")
        token = input("> ").strip() # getpass doesn't allow copy-paste
    pk = Client(token)

    while True:
        try:
            print((
                "Enter the ID of the system whose members to list, "
                "empty if getting one's own members."
            ))
            id = input("> ").strip().lower()
            if not id:
                asyncio.get_event_loop().run_until_complete(main(pk, None))
            elif len(id) == 5 and all(c.isalpha() for c in id):
                asyncio.get_event_loop().run_until_complete(main(pk, id))
            else:
                print("That's not a valid system ID.")
        except KeyboardInterrupt:
            print("^C")
            break
        except:
            logging.exception("An exception occurred")

if __name__ == "__main__":
    run()



from collections import namedtuple

from .v1.client import Client
from .v1.models import (
    Member,
    System,
    ProxyTag,
    ProxyTags,
    Switch,
    Privacy,
    Timestamp,
    Color,
    Birthday,
    Timezone,
    Message,
)
from .v1 import errors

VersionInfo = namedtuple("VersionInfo", "major minor build")
version_info = VersionInfo(
    major=0,
    minor=0,
    build=6,
)

__title__ = "pluralkit"
__author__ = "Madison Landry, Alyx Warner"
__copyright__ = "Copyright 2021-present Madison Landry, Alyx Warner"
__version__ = f"{version_info.major}.{version_info.minor}.{version_info.build}"



import logging
import sys
sys.path.append("..")

import asyncio

from pluralkit import Client

async def main(pk, id):
    system = await pk.get_system(id)
    print(f"{system.name} (`{system.id}`)")

def run(token: str=None):
    if not token:
        print("Enter the token, if desired, otherwise press Enter.")
        token = input("> ").strip() # getpass doesn't allow copy-paste
    pk = Client(token)
    
    while True:
        try:
            print("Enter the ID of the system to get, empty if getting one's own system.")
            id = input("> ").strip().lower()
            if not id:
                asyncio.get_event_loop().run_until_complete(main(pk, None))
            elif len(id) == 5 and all(c.isalpha() for c in id):
                asyncio.get_event_loop().run_until_complete(main(pk, id))
            else:
                print("That's not a valid system ID.")
        except KeyboardInterrupt:
            print("^C")
            break
        except:
            logging.exception("An exception occurred")

if __name__ == "__main__":
    run()



__all__ = (
    "PluralKitException",
    "AuthorizationError",
    "SystemNotFound",
    "DiscordUserNotFound",
    "AccessForbidden",
    "InvalidKwarg",
    "InvalidColor",
    "InvalidBirthday",
    "HTTPError"
)

class PluralKitException(Exception):
    """Base exception class for pk.py
    """
    ...

class AuthorizationError(PluralKitException):
    """Thrown when the authorization token passed to PluralKit's API is invalid (or missing).
    """
    def __init__(self):
        super().__init__((
            "System token seems to be missing or invalid. Can you check that you entered it in "
            "correctly?"
        ))

class SystemNotFound(PluralKitException):
    """Thrown when the system ID is apparently not in PluralKit's database.
    """
    def __init__(self, id):
        super().__init__(
            f"System with the given ID (`{id}`) was not found."
        )

class MemberNotFound(PluralKitException):
    """Thrown when the system ID is apparently not in PluralKit's database.
    """
    def __init__(self, id):
        super().__init__(
            f"Member with the given ID (`{id}`) was not found."
        )

class DiscordUserNotFound(PluralKitException):
    """Thrown when the Discord user ID is apparently not associated with a PluralKit system.
    """
    def __init__(self, id):
        super().__init__(
            f"The given user ID (`{id}`) does not seem to correspond to any systems."
        )

class AccessForbidden(PluralKitException):
    """Thrown when a system's list of members is private.
    """
    def __init__(self):
        super().__init__((
            "The system's member list is private and the client authorization token is either "
            "missing, invalid, or does not correspond to the system."
        ))

class InvalidKwarg(PluralKitException):
    """
    Thrown when an invalid field is passed in a POST or PATCH request
    """
    def __init__(self, key):
        super().__init__(
            f"A keyworded argument was passed that will not be accepted by the server: `{key}`"
        )

class InvalidColor(PluralKitException):
    """
    Thrown when an invalid color is passed in a POST or PATCH request
    """
    def __init__(self, color):
        super().__init__(
                f"Given value is not a string or Color object: `{color}`"
            )

class InvalidBirthday(PluralKitException):
    """
    Thrown when an invalid string is passed for the "Birthday" field of a member object. 
    (Must be yyyy-mm-dd)
    """
    def __init__(self, string):
        super().__init__(
            f"`{string}` is not a valid yyyy-mm-dd date or datetime.datetime object"
        )

class HTTPError(PluralKitException):
    """Todo.
    """
    def __init__(self, error):
        super().__init__(
            f"An error was raised: {error}"
        )



import asyncio
import functools

def func(f):
   """Makes an asynchronous `Client` function synchronous if `Client.async_mode` == ``False``.
   """
   @functools.wraps(f)
   def wrapper(*args, **kwargs):
      self = args[0]

      # save an asynchronous copy! if not yet saved
      async_name = "_" + f.__name__
      try:
         getattr(self, async_name)
      except AttributeError:
         setattr(self, async_name, f)

      awaitable = f(*args, **kwargs)
      if self.async_mode:
         return awaitable

      loop = asyncio.get_event_loop()
      result = loop.run_until_complete(awaitable)
      return result

   return wrapper

def iter(f):
   """Makes an asynchronous `Client` iterator synchronous if `Client.async_mode` == ``False``.
   """
   @functools.wraps(f)
   def wrapper(*args, **kwargs):
      self = args[0]

      # save an asynchronous copy! if not yet saved
      async_name = "_" + f.__name__
      try:
         getattr(self, async_name)
      except AttributeError:
         setattr(self, async_name, f)

      awaitable = f(*args, **kwargs)
      if self.async_mode:
         return awaitable

      async def flatten(awaitable):
         return [item async for item in awaitable]

      loop = asyncio.get_event_loop()
      result = loop.run_until_complete(flatten(awaitable))
      return result

   return wrapper



from datetime import datetime, timedelta, tzinfo
from enum import Enum
import string
from typing import (
    Any,
    Union, Optional,
    Tuple, List, Set, Sequence, Dict,
    Generator,
)

import pytz
import colour
from .errors import *

class Privacy(Enum):
    """Represents the privacies accepted by PluralKit.
    """
    PUBLIC = "public"
    PRIVATE = "private"
    UNKNOWN = None # legacy, effectively resets privacy to "public"

class Color(colour.Color):
    """Represents a color.

    This class is initialized in the same way that a `colour.Color`_ object is. It may also take a
    `colour.Color`_ object directly.

    .. _`colour.Color`: https://pypi.org/project/colour/#instantiation
    """
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], colour.Color):
            super().__init__(args[0].hex_l)
        else:
            super().__init__(*args, **kwargs)

    @staticmethod
    def parse(c):
        """Takes in a `Color`, `colour.Color`_, or str and converts to `Color` as
        needed.

        Args:
            color (Union[Color,colour.Color,str,None]): The color, represented as a `Color`,
                `colour.Color`_ or `str`. If a string, may either be in the format as expected by
                PluralKit's API internally (e.g. ``00ffff``) or a color string that can be taken by
                a Color object (e.g. ``cyan``).

        Returns:
            Optional[Color]: The `Color` object, or ``None`` if input is None.

        Raises:
            TypeError: If the given argument is neither a `Color`, `colour.Color`_, or `str`.

        .. _`colour.Color`: https://pypi.org/project/colour/#instantiation
        """
        if c is None: return None

        if isinstance(c, colour.Color):
            return c

        if isinstance(c, str):
            if len(c) == 6 and set(c).issubset(set(string.hexdigits)):
                return Color.from_json(c)
            else:
                return Color(c)

        raise TypeError(
            f"Argument `c` must be of type colour.Color or str; received {type(c)=}."
        )

    @staticmethod
    def from_json(color: str):
        """Takes in a string (as returned by the API) and returns the `Color`.
        """
        return Color.__new__(Color, f"#{color}") # mypy says this has too many args?

    def json(self):
        """Returns the hex of the `Color` sans the ``#`` symbol.

        Example: ``Color("magenta").json()`` would return ``"ff00ff"``.
        """
        return self.hex_l[1:]

class Timezone:
    """Represents a tzdb time zone.

    This class is initialized in the same way that `pytz.timezone`_ initializes `tzinfo` objects.
    It may also take a `tzinfo` object directly.

    Hint:
        `Here is a link to a list of tz database time zones`_

    Args:
        tz (Union[str,tzinfo]): The timezone, either as a string or as a `tzinfo` (e.g. from
            `pytz.timezone`_).

    .. _`pytz.timezone`: http://pytz.sourceforge.net/
    .. _`Here is a link to a list of tz database time zones`:
        https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """
    def __init__(self, *args, **kwargs):
        if len(args) != 1 or len(kwargs) != 0:
            raise TypeError(
                f"Timezone is initialized with exactly one positional argument `tz`; " \
                f"received {len(args)=} and {len(kwargs)=}"
            )
        if isinstance(args[0], tzinfo):
            self.tz = args[0]
        else:
            self.tz = pytz.timezone(args[0])

    def __repr__(self):
        return f"{self.__class__.__name__}<{self.zone}>"

    @property
    def zone(self):
        return self.tz.zone

    @staticmethod
    def parse(tz):
        """Takes in a `Timezone`, `tzinfo`, or `str` and converts to `Timezone` as needed.

        Args:
            tz (Union[Timezone,tzinfo,str]): The timezone, represented as a
                `Timezone`, `tzinfo`, or `str`.

        Raises:
            TypeError: If given argument is neither a `Timezone`, `tzinfo`, nor `str`.
        """
        if isinstance(tz, Timezone):
            return tz

        if isinstance(tz, (tzinfo, str)):
            return Timezone(tz)

        raise TypeError(
            f"Argument `tz` must be of type Timezone, tzinfo, or str; " \
            f"received {type(tz)=}."
        )

    @staticmethod
    def from_json(tz: str):
        """Takes in a string (as returned by the API) and returns the Timezone.

        Args:
            tz: The time zone as stored in the PluralKit API internally.

        Returns:
            Timezone: The corresponding `Timezone` object.
        """
        return Timezone(tz)

    def json(self):
        """Returns the string representation of this timezone as expected by the API.
        """
        return self.zone
    
class Timestamp:
    """Represents a PluralKit UTC timestamp.

    This class works by wrapping around a `datetime` object. Use ts.datetime to access it, for any
    `Timestamp` ts.

    This class may be initialized in the same way that a `datetime` object is. It may also take a
    `datetime` object directly.
    """
    def __init__(self, dt: Optional[datetime]=None, *,
        year: Optional[int]=None,
        month: Optional[int]=None,
        day: Optional[int]=None,
        hour: int=0,
        minute: int=0,
        second: int=0,
        microsecond: int=0
    ):
        if dt is None and any(arg is None for arg in (year, month, day)):
            raise TypeError(
                f"{self.__class__.__name__} is missing required arguments. Either provide a " \
                f"datetime.datetime via the first positional argument, or provide the year, " \
                f"month, and day through the respective keyword arguments."
            )

        if dt is not None:
            if dt.tzinfo is not None:
                self.datetime = dt.astimezone(pytz.utc)
            else:
                self.datetime = dt.replace(tzinfo=pytz.utc)

        else:
            # mypy complains here
            self.datetime = datetime(year, month, day, hour, minute, second, microsecond)
            self.datetime = self.datetime.replace(tzinfo=pytz.utc)

    def __repr__(self):
        return f"{self.__class__.__name__}<{self.json()}>"

    def __str__(self):
        return (
            f"{self.year:04d}-{self.month:02d}-{self.day:02d} "
            f"{self.hour:02d}:{self.minute:02d}:{self.second:02d} UTC"
        )

    @property
    def year(self):
        return self.datetime.year

    @year.setter
    def year(self, value):
        self.datetime = self.datetime.replace(year=value)

    @property
    def month(self):
        return self.datetime.month

    @month.setter
    def month(self, value):
        self.datetime = self.datetime.replace(month=value)

    @property
    def day(self):
        return self.datetime.day

    @day.setter
    def day(self, value):
        self.datetime = self.datetime.replace(day=value)

    @property
    def hour(self):
        return self.datetime.hour

    @hour.setter
    def hour(self, value):
        self.datetime = self.datetime.replace(hour=value)

    @property
    def minute(self):
        return self.datetime.minute

    @minute.setter
    def minute(self, value):
        self.datetime = self.datetime.replace(minute=value)

    @property
    def second(self):
        return self.datetime.second

    @second.setter
    def second(self, value):
        self.datetime = self.datetime.replace(second=value)

    @property
    def microsecond(self):
        return self.datetime.microsecond

    @microsecond.setter
    def microsecond(self, value):
        self.datetime = self.datetime.replace(microsecond=value)

    @staticmethod
    def parse(ts):
        """Takes in a `Timestamp`, `datetime`_, or `str`, converts to `Timestamp` as needed.

        Args:
            ts (Union[Timestamp,datetime,str]): The timestamp, represented as a `Timestamp`,
                `datetime`_, or `str`.

        Returns:
            Timestamp: The `Timestamp` object.

        Raises:
            TypeError: If given argument is neither a `Timestamp`, `datetime`_, or `str`.

        .. _`datetime`: https://docs.python.org/3/library/datetime.html#datetime-objects
        """
        if isinstance(ts, Timestamp):
            return ts

        if isinstance(ts, datetime):
            return Timestamp(ts)

        if isinstance(ts, str):
            return Timestamp.from_json(ts)

        raise TypeError(
            f"Argument `ts` must be of type Timestamp, datetime.datetime, or str; " \
            f"received {type(ts)=}."
        )

    @staticmethod
    def from_json(bd: str):
        """Takes in a string (as returned by the API) and returns the corresponding `Timestamp`.

        Args:
            bd: The ``{year}-{month}-{day}T{hour}:{minute}:{second}.{microsecond}Z`` formatted
                string representing a PluralKit API timestamp.

        Returns:
            Timestamp: The corresponding `Timestamp` object.
        """
        return Timestamp(datetime.strptime(bd, r"%Y-%m-%dT%H:%M:%S.%fZ"))

    def json(self) -> str:
        """Convert this timestamp to the ISO 8601 format that PluralKit uses internally.
        """
        return (
            f"{self.year:04d}-{self.month:02d}-{self.day:02d}"
            f"T{self.hour:02d}:{self.minute:02d}:{self.second:02d}.{self.microsecond:06d}Z"
        )

class Birthday(Timestamp):
    """Represents a birthday.
    """
    def __str__(self):
        if self.hidden_year:
            return self.datetime.strftime("%b %d")
        else:
            return self.datetime.strftime("%b %d, ") + f"{self.year:04d}"

    @property
    def hidden_year(self) -> bool:
        """Whether this birthday's year is hidden.

        If set to ``False``, sets the birthday's year to ``0001``, `which internally represents a
        hidden year in PluralKit's API`_.

        .. _`which internally represents a hidden year in PluralKit's API`:
            https://pluralkit.me/api/#member-model
        """
        return self.year in (1, 4)

    @hidden_year.setter
    def hidden_year(self, value: bool):
        self.year = 1

    @staticmethod
    def parse(bd):
        """Takes in a `Birthday`, `datetime`_, or str, converts to `Birthday` as needed.

        Args:
            bd (Union[Birthday,datetime,str,None]): The birthday, represented as a `Birthday`,
                `datetime`_, or str.

        Returns:
            Optional[Birthday]: The `Birthday` object, or ``None`` if input is ``None``.

        Raises:
            TypeError: If given argument is neither a `Birthday`, `datetime`_, or str.

        .. _`datetime`: https://docs.python.org/3/library/datetime.html#datetime-objects
        """
        if bd is None: return None

        if isinstance(bd, Birthday):
            return bd

        if isinstance(bd, datetime):
            return Birthday(bd)

        if isinstance(bd, str):
            return Birthday.from_json(bd)

        raise TypeError(
            f"Argument `bd` must be None or of type Birthday, datetime.datetime, or str; " \
            f"received {type(bd)=}."
        )

    @staticmethod
    def from_json(bd: str):
        """Takes in a string (as returned by the API) and returns the `Birthday`.

        Args:
            bd: The ``YYYY-MM-DD`` formatted string representing the birthdate.

        Returns:
            Birthday: The corresponding birthday.
        """
        return Birthday(datetime.strptime(bd, r"%Y-%m-%d"))

    def json(self) -> str:
        """Returns the ``YYYY-MM-DD`` formatted birthdate.
        """
        return f"{self.year:04d}-{self.month:02d}-{self.day:02d}"

class ProxyTag:
    """Represents a single PluralKit proxy tag.

    Args:
        prefix: Prefix that will enclose proxied messages.
        suffix: Suffix that will enclose proxied messages.

    Important:
        At least one of the ``suffix`` or ``prefix`` arguments must be passed.
    
    Attributes:
        prefix (Optional[str]): Prefix that will enclose proxied messages.
        suffix (Optional[str]): Suffix that will enclose proxied messages.
    """
    def __init__(self, *,
        prefix: Optional[str]=None,
        suffix: Optional[str]=None,
    ):
        assert prefix or suffix, \
            "A valid proxy tag must have at least one of the prefix or suffix defined."
        self.prefix = prefix
        self.suffix = suffix

    @staticmethod
    def from_json(proxy_tag: Dict[str,str]):
        """Static method to convert a proxy tag `dict` to a `ProxyTag`.

        Args:
            proxy_tag: Dictionary representing a proxy tag. Must have at least one of ``prefix`` or
                ``suffix`` as keys.

        Returns:
            ProxyTag: The corresponding `ProxyTag` object.
        """
        return ProxyTag(
            prefix=proxy_tag.get("prefix"),
            suffix=proxy_tag.get("suffix")
        )

    def __repr__(self):
        prefix = "" if not self.prefix else f"prefix={repr(self.prefix)}"
        suffix = "" if not self.suffix else f"suffix={repr(self.suffix)}"
        attrs = ",".join(a for a in (prefix, suffix) if a)
        return (
            f"{self.__class__.__name__}({attrs})"
        )

    def match(self, message: str) -> bool:
        """Determine if a given message would be proxied under this proxy tag.
        
        Args:
            message: Message to parse.
        """
        message = message.strip()
        return (True if not self.prefix else message.startswith(self.prefix)) \
            and (True if not self.suffix else message.endswith(self.suffix))

    def json(self) -> Dict[str,Optional[str]]:
        """Return the JSON object representing this proxy tag as a Python `dict`.
        """
        return {
            "prefix": self.prefix,
            "suffix": self.suffix,
        }

class ProxyTags:
    """Represents a set of PluralKit proxy tags.

    Hint:
        ProxyTags objects can be iterated or indexed to yield its underlying `ProxyTag` objects.    
    
    Args:
        proxy_tags: A sequence of `ProxyTag` objects.
    """
    def __init__(self, proxy_tags: Optional[Generator[ProxyTag,None,None]]=None):
        self._proxy_tags: Tuple[ProxyTag,...]
        if proxy_tags is None:
            self._proxy_tags = tuple()
        else:
            self._proxy_tags = tuple(proxy_tags)

    def __repr__(self):
        return f"{self.__class__.__name__}<{len(self._proxy_tags)}>"

    def __iter__(self):
        for proxy_tag in self._proxy_tags:
            yield proxy_tag

    def __getitem__(self, index):
        return self._proxy_tags[index]

    @staticmethod
    def from_json(proxy_tags: Sequence[Dict[str,str]]):
        """Static method to convert a list of proxy tags `dict`s to a `ProxyTags` object.

        Args:
            proxy_tags: Sequence of Python dictionaries, each representing a proxy tag.

        Returns:
            ProxyTags: The corresponding `ProxyTags` object.
        """
        return ProxyTags(ProxyTag.from_json(proxy_tag) for proxy_tag in proxy_tags)

    def match(self, message: str) -> bool:
        """Determine if a given message would be proxied under this set of proxy tags.
        
        Args:
            message: Message to parse.
        """
        return any(proxy_tag.match(message) for proxy_tag in self)

    def json(self) -> List[Dict[str,str]]:
        """Return the JSON object representing this proxy tag as a list of Python `dict`.
        """
        return [proxy_tag.json() for proxy_tag in self]

class System:
    """Represents a PluralKit system.

    Args:
        id: The system's five-character lowercase ID.
        created: The system's creation date. May be a string formatted as
            ``{year}-{month}-{day}T{hour}:{minute}:{second}.{microsecond}Z`` (ISO 8601 format), a
            `Timestamp`, or a `datetime`_.
        name: The name of the system. Default ``None``.
        description: The description of the system. Default ``None``.
        tag: The system's tag appended to display names. Default ``None``.
        avatar_url: The system's avatar URL. Default ``None``.
        tz: The system's tzdb timezone. May be a `Timezone`, `tzinfo`, or `str`. Default is
            ``"UTC"``.
        description_privacy: The system's description privacy, either `~Privacy.PUBLIC` or
            `~Privacy.PRIVATE`. Default is public.
        member_list_privacy: The system's member list privacy, either `~Privacy.PUBLIC` or
            `~Privacy.PRIVATE`. Default is public.
        front_privacy: The system's fronting privacy, either `~Privacy.PUBLIC` or
            `~Privacy.PRIVATE`. Default is public.
        front_history_privacy: The system's fronting history privacy, either `~Privacy.PUBLIC` or
            `~Privacy.PRIVATE`. Default is public.

    Attributes:
        id (str): The system's five-character lowercase ID.
        name (Optional[str]): The name of the system.
        description (Optional[str]): The description of the system.
        tag (Optional[str]): The system's tag appended to display names.
        avatar_url (Optional[str]): The system's avatar URL.
        tz (Timezone): The system's tzdb timezone.
        created (Timestamp): The system's timestamp at creation.
        description_privacy (Privacy): The system's description privacy.
        member_list_privacy (Privacy): The system's member list privacy.
        front_privacy (Privacy): The system's fronting privacy.
        front_history_privacy (Privacy): The system's fronting history privacy.

    .. _`datetime`: https://docs.python.org/3/library/datetime.html#datetime-objects
    """

    def __init__(self, *,
        id: str,
        created: Union[Timestamp,datetime,str],
        name: Optional[str]=None,
        description: Optional[str]=None,
        tag: Optional[str]=None,
        avatar_url: Optional[str]=None,
        tz: Union[Timezone,tzinfo,str]="UTC",
        description_privacy: Union[Privacy,str]=Privacy.PUBLIC,
        member_list_privacy: Union[Privacy,str]=Privacy.PUBLIC,
        front_privacy: Union[Privacy,str]=Privacy.PUBLIC,
        front_history_privacy: Union[Privacy,str]=Privacy.PUBLIC
    ):
        self.id = id
        self.name = name

        self.description = description
        self.tag = tag
        self.avatar_url = avatar_url

        self.created = Timestamp.parse(created)
        self.tz = Timezone.parse(tz)

        self.description_privacy = Privacy(description_privacy)
        self.member_list_privacy = Privacy(member_list_privacy)
        self.front_privacy = Privacy(front_privacy)
        self.front_history_privacy = Privacy(front_history_privacy)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.id}')"

    def __str__(self):
        return self.id

    @staticmethod
    def from_json(system: Dict[str,Any]):
        """Static method to convert a system `dict` to a `System` object.

        Args:
            system: Dictionary representing a system, e.g. one received directly from the API. Must
                have a value for the ``id`` and ``created`` attributes.

        Returns:
            System: The corresponding `System` object.
        """
        return System(
            id=system["id"],
            name=system.get("name"),
            description=system.get("description"),
            tag=system.get("tag"),
            avatar_url=system.get("avatar_url"),
            tz=system.get("tz", "UTC"),
            created=system["created"],
            description_privacy=system.get("description_privacy", "public"),
            member_list_privacy=system.get("member_list_privacy", "public"),
            front_privacy=system.get("front_privacy", "public"),
            front_history_privacy=system.get("front_history_privacy", "public")
        )

    def json(self) -> Dict[str,Any]:
        """Return Python `dict` representing this system.
        """
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "tag": self.tag,
            "avatar_url": self.avatar_url,
            "tz": self.tz.json(),
            "created": self.created.json(),
            "description_privacy": self.description_privacy.value,
            "member_list_privacy": self.member_list_privacy.value,
            "front_privacy": self.front_privacy.value,
            "front_history_privacy": self.front_history_privacy.value
        }
    
class Member:
    """Represents a PluralKit system member.

    Args:
        id: The member's five-letter lowercase ID.
        name: The member's name.
        created: The member's creation date. May be a string formatted as
            ``{year}-{month}-{day}T{hour}:{minute}:{second}.{microsecond}Z`` (ISO 8601 format), a
            `Timestamp`, or a `datetime`_.
        name_privacy: The member's name privacy, either `~Privacy.PUBLIC` or `~Privacy.PRIVATE`.
            Default is public.
        display_name: The member's display name. Default is ``None``.
        description: The member's description. Default is ``None``.
        description_privacy: The member's description privacy, either `~Privacy.PUBLIC` or
            `~Privacy.PRIVATE`. Default is public.
        color: The member's color. Default is ``None``.
        birthday: The member's birthdate. May be a string formatted as ``{year}-{month}-{day}``, a
            `Timestamp`, or a `datetime`_. Default is ``None``.
        birthday_privacy: The member's birthdate privacy, either `~Privacy.PUBLIC` or
            `~Privacy.PRIVATE`. Default is public.
        pronouns: The member's pronouns. Default is ``None``.
        pronoun_privacy: The member's pronouns privacy, either `~Privacy.PUBLIC` or
            `~Privacy.PRIVATE`. Default is public.
        avatar_url: The member's avatar URL.
        avatar_privacy: The member's avatar privacy, either `~Privacy.PUBLIC` or
            `~Privacy.PRIVATE`. Default is public.
        keep_proxy: Whether the member's proxy tags remain in the proxied message (``True``) or not
            (``False``). Default is ``False``.
        metadata_privacy: The member's metadata (eg. creation timestamp, message count, etc.)
            privacy. Must be either `~Privacy.PUBLIC` or `~Privacy.PRIVATE`. Default is public.
        proxy_tags: The member's proxy tags. Default is an empty set of proxy tags.
        visibility: The visibility privacy setting of the member, either `~Privacy.PUBLIC` or
            `~Privacy.PRIVATE`. Default is public.

    Attributes:
        id (str): The member's five-letter lowercase ID.
        name (str): The member's name.
        created (Timestamp): The member's creation date.
        name_privacy (Privacy): The member's name privacy.
        display_name (Optional[str]): The member's display name.
        description (Optional[str]): The member's description.
        description_privacy (Privacy): The member's description privacy.
        color (Optional[Color]): The member's color.
        birthday (Optional[Birthday]): The member's birthdate.
        birthday_privacy (Privacy): The member's birthday privacy.
        pronouns (Optional[str]): The member's pronouns.
        pronoun_privacy (Privacy): The member's pronouns privacy.
        avatar_url (Optional[str]): The member's avatar URL.
        avatar_privacy (Privacy): The member's avatar privacy.
        keep_proxy (bool): Whether the member's proxy tags remain in the proxied message (``True``)
            or not (``False``).
        metadata_privacy (Privacy): The member's metadata (eg. creation timestamp, message count,
            etc.) privacy.
        proxy_tags (ProxyTags): The member's proxy tags.
        visibility (Privacy): The visibility privacy setting of the member.

    .. _`datetime`: https://docs.python.org/3/library/datetime.html#datetime-objects
    """

    def __init__(self, *,
        id: str,
        name: str,
        created: Union[Timestamp,datetime,str],
        name_privacy: Union[Privacy,str]=Privacy.PUBLIC,
        display_name: Optional[str]=None,
        description: Optional[str]=None,
        description_privacy: Union[Privacy,str]=Privacy.PUBLIC,
        color: Union[Color,str,None]=None,
        birthday: Union[Birthday,datetime,str,None]=None,
        birthday_privacy: Union[Privacy,str]=Privacy.PUBLIC,
        pronouns: Optional[str]=None,
        pronoun_privacy: Union[Privacy,str]=Privacy.PUBLIC,
        avatar_url: Optional[str]=None,
        avatar_privacy: Union[Privacy,str]=Privacy.PUBLIC,
        keep_proxy: bool=False,
        metadata_privacy: Union[Privacy,str]=Privacy.PUBLIC,
        proxy_tags: Optional[ProxyTags]=None,
        visibility: Union[Privacy,str]=Privacy.PUBLIC
    ):
        self.id = id
        self.name = name

        self.created = Timestamp.parse(created)
        self.birthday = Birthday.parse(birthday)
        self.color = Color.parse(color)

        self.display_name = display_name
        self.description = description
        self.pronouns = pronouns
        self.avatar_url = avatar_url
        self.keep_proxy = keep_proxy

        if proxy_tags is None:
            self.proxy_tags = ProxyTags()
        else:
            self.proxy_tags = proxy_tags

        self.name_privacy = Privacy(name_privacy)
        self.description_privacy = Privacy(description_privacy)
        self.birthday_privacy = Privacy(birthday_privacy)
        self.pronoun_privacy = Privacy(pronoun_privacy)
        self.avatar_privacy = Privacy(avatar_privacy)
        self.metadata_privacy = Privacy(metadata_privacy)
        self.visibility = Privacy(visibility)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.id}')"

    def __str__(self):
        return self.id

    @staticmethod
    def from_json(member: Dict[str,Any]):
        """Static method to convert a member `dict` to a `Member` object.

        Args:
            member: Dictionary representing a system, e.g. one received directly from the API. Must
            have a value for the ``id`` and ``created`` attributes.

        Returns:
            Member: The corresponding `Member` object.
        """
        if not "proxy_tags" in member:
            proxy_tags = ProxyTags()
        else:
            proxy_tags = ProxyTags.from_json(member["proxy_tags"])
        return Member(
            id=member["id"],
            name=member["name"],
            name_privacy=member.get("name_privacy", "public"),
            created=member["created"],
            display_name=member.get("display_name"),
            description=member.get("description"),
            description_privacy=member.get("description_privacy", "public"),
            color=member.get("color"),
            birthday=member.get("birthday"),
            birthday_privacy=member.get("birthday_privacy", "public"),
            pronouns=member.get("pronouns", "public"),
            pronoun_privacy=member.get("pronoun_privacy", "public"),
            avatar_url=member.get("avatar_url"),
            avatar_privacy=member.get("avatar_privacy", "public"),
            keep_proxy=member.get("keep_proxy", False),
            metadata_privacy=member.get("metadata_privacy", "public"),
            proxy_tags=proxy_tags,
            visibility=member.get("visibility", "public"),
        )

    def json(self) -> Dict[str,Any]:
        """Generate the Python `dict` representing this member.
        """
        return {
            "id": self.id,
            "name": self.name,
            "name_privacy": self.name_privacy.value,
            "created": self.created.json(),
            "display_name": self.display_name,
            "description": self.description,
            "description_privacy": self.description_privacy.value,
            "color": self.color.json(),
            "birthday": self.birthday.json(),
            "birthday_privacy": self.birthday_privacy.value,
            "pronouns": self.pronouns,
            "pronoun_privacy": self.pronoun_privacy.value,
            "avatar_url": self.avatar_url,
            "avatar_privacy": self.avatar_privacy.value,
            "keep_proxy": self.keep_proxy,
            "metadata_privacy": self.metadata_privacy.value,
            "proxy_tags": self.proxy_tags.json(),
            "visibility": self.visibility.value,
        }

class Switch:
    """Represents a switch event.

    Args:
        timestamp: Timestamp of the switch. May be a string formatted as
            ``{year}-{month}-{day}T{hour}:{minute}:{second}.{microsecond}Z`` (ISO 8601 format), a
            `Timestamp`, or a `datetime`_.
        members: Members involved. May be a list of the five-letter member IDs as strings, or a
            list of `Member` models, though cannot be mixed.

    Attributes:
        timestamp (Timestamp): Timestamp of the switch.
        members (Union[Sequence[str],Sequence[Member]]): Members involved.

    .. _`datetime`: https://docs.python.org/3/library/datetime.html#datetime-objects
    """
    def __init__(self, *,
        timestamp: Union[Timestamp,str],
        members: Union[Sequence[str],Sequence[Member]]
    ):
        self.timestamp = Timestamp.parse(timestamp)

        if members is None or len(members) == 0:
            self.members = []
        else:
            self.members = [member for member in members]

    def __repr__(self):
        return f"{self.__class__.__name__}<{self.timestamp}>"

    @staticmethod
    def from_json(switch: Dict[str,str]):
        """Static method to convert a switch `dict` to a `Switch` object.

        Args:
            switch: Dictionary representing a switch, e.g. one received directly from the API. Must
            have a value for the ``members`` and ``timestamp`` attributes. See this class's
            initializer documentation for what format those are expected to be in.

        Returns:
            Switch: The corresponding `Switch` object.
        """
        return Switch(
            timestamp=switch["timestamp"],
            members=switch["members"]
        )

    def json(self) -> Dict[str,Any]:
        """Return Python `dict` representing this switch.
        """
        return {
            "timestamp": self.timestamp.json(),
            "members": self.members
        }

class Message:
    """Represents a proxied message.

    Args:
        timestamp: Timestamp of the message. May be a string for
            atted as ``{year}-{month}-{day}T{hour}:{minute}:{second}.{microsecond}Z`` (ISO 8601
            format), a `Timestamp`, or a `datetime`_.
        id: The ID of the Discord message sent by the webhook.
        original: The ID of the (presumably deleted) original Discord message sent by the account.
        sender: The user ID of the account that sent the message.
        channel: The ID of the channel the message was sent to.
        system: The system that proxied the message.
        member: The member that proxied the message.

    Attributes:
        timestamp (Timestamp): Timestamp of the message.
        id (int): The ID of the Discord message sent by the webhook.
        original (int): The ID of the (presumably deleted) original Discord message sent by the
            account.
        sender (int): The user ID of the account that sent the message.
        channel (int): The ID of the channel the message was sent to.
        system (System): The system that proxied the message.
        member (Member): The member that proxied the message.

    .. _`datetime`: https://docs.python.org/3/library/datetime.html#datetime-objects
    """
    def __init__(self, *,
        timestamp: Union[Timestamp,datetime,str],
        id: Union[int,str],
        original: Union[int,str],
        sender: Union[int,str],
        channel: Union[int,str],
        system: System,
        member: Member
    ):
        self.id = int(id)
        self.original = int(original)
        self.sender = int(sender)
        self.channel = int(channel)
        self.system = system
        self.member = member

        self.timestamp = Timestamp.parse(timestamp)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id})"

    @staticmethod
    def from_json(message: Dict[str,Any]):
        """Static method to convert a message `dict` to a `Message` object.

        Args:
            message: Dictionary representing a message, e.g. one received directly from the API.
            Must have a value the same attributes as required by the `Message` initializer.

        Returns:
            Message: The corresponding `Message` object.
        """
        return Message(
            timestamp=message["timestamp"],
            id=message["id"],
            original=message["original"],
            sender=message["sender"],
            channel=message["channel"],
            system=System.from_json(message["system"]),
            member=Member.from_json(message["member"])
        )

    def json(self) -> Dict[str,Any]:
        """Return Python `dict` representing this message.
        """
        return {
            "timestamp": self.timestamp.json(),
            "id": str(self.id),
            "original": str(self.original),
            "sender": str(self.sender),
            "channel": str(self.channel),
            "system": self.system.json(),
            "member": self.member.json()
        }






from setuptools import setup, find_packages

from pluralkit import __version__

# Thanks to Mark Smith (@Judy2k)'s very helpful talk about PyPI
# https://youtu.be/GIF3LaRqgXo?t=297

with open("README.md", "r") as fh:
   long_description = fh.read()

setup(
   name="pluralkit",
   version=__version__,
   description="Python wrapper for PluralKit's API.",
   license="MIT",
   packages=find_packages(),
   classifiers=[ # https://pypi.org/classifiers/
      "Development Status :: 3 - Alpha",
      "Programming Language :: Python :: 3.6", # f-strings
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
      "Intended Audience :: Developers",
      "Natural Language :: English",
      "Topic :: Internet",
      "Topic :: Software Development :: Libraries",
      "Topic :: Software Development :: Libraries :: Python Modules",
      "Topic :: Utilities",
   ],
   long_description=long_description,
   long_description_content_type="text/markdown",
   install_requires=[
      "httpx==0.18.2", # https://www.python-httpx.org/
      "colour>=0.1",
      "pytz>=2021",
   ],
   extras_require = {
      "dev": [
         "Sphinx==3.5.4", # documentation!
         "sphinx-autodoc-typehints==1.12.0", # better sphinx parsing
         "sphinx-book-theme==0.1.0", # Book theme
         "mypy==0.910", # type checking
         "types-pytz==2021.1.0", # type checking with pytz
         "pytest==6.2.4", # testing module
         "check-manifest==0.46", # creating MANIFEST.in
         "twine==3.4.1", # for uploading to PyPI
         "wheel>=0.36.2", # for building wheels
      ]
   },
   python_requires=">=3.6.0",
   url="https://github.com/almonds0166/pluralkit.py",
   author="Madison Landry, Alyx Warner",
   author_email="pkpy@mit.edu",
   project_urls={
     "Documentation": "https://pluralkit.readthedocs.io/en/latest/",
     "Issue tracker": "https://github.com/almonds0166/pluralkit.py/issues",
   },
)